
/********************************************************************
 *
 *	TITLE:	   x16
 *
 *	DATE:	   05-26-82
 *
 *	ABSTRACT:  .
 *
 *	LANGUAGE DEPENDENCIES:
 *
 *		C86     VERSION 1.0
 *
 *********************************************************************/
#include <sys/types.h>
#include <curses.h>
#include <signal.h>
#include "common.def"

	/*
	 *  Public variables
	 */

int  go[2];
int  synch[2];
int  h_update = 120;
bool idleps = FALSE;
int  speed = CRUISE;
int all_io = FALSE;
int all_math = FALSE;
int num_8087 = 0;

struct {
	int       number;
	unsigned  min;
	unsigned  max;
	char name[NAMEMAX];
} paths[P_MAX];

	/*
	 *  External Variables
	 */

extern int errno;
extern int sys_nerr;
extern char *sys_errlist[];
extern char *index();
extern long atol();
extern long time();

	/*
	 * Static variables
	 */

static long cal = 0; 
static int  c_update = 2;
static int  pi = 0;
static bool cur_on = FALSE;
static int forkret;
static char *fpram[2] = {
	"fill",
	""
};

static char *params[8] = {
	"io",
	"math",
	"55",
	"fast",
	"8087",
	"calibrate",
	"idle",
	""
};

static char go_msg[] = {"123456789012345678"};

/********************************************************************
 *
 *	TITLE:	   quit
 *
 *	CALLING SEQUENCE:
 *
 *		   quit();
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
quit()
{

	endwin();

	exit(0);

}

/********************************************************************
 *
 *	TITLE:	   err_msg
 *
 *	CALLING SEQUENCE:
 *
 *		   err_msg( msg1, msg2, sys_err );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
err_msg( msg1, msg2, sys_err )
char *msg1, *msg2;
bool sys_err;
{

	printf( "%s%s", msg1, msg2 );
	if ( sys_err ) {
		printf( ": " );
		if ( errno <= sys_nerr )
			printf( "%s", sys_errlist[errno] );
		else
			printf( "unknown sys error" );
	}
	printf( "\n" );

	if ( cur_on )
		kill( 0, SIGINT );
	else
		exit(1);	
}

/********************************************************************
 *
 *	TITLE:	   find
 *
 *	CALLING SEQUENCE:
 *
 *		   newsp = find( sp );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
char *find( sp )
char *sp;
{
char *nextsp;

	if ( *sp == '\0' )
		return( NULL );

	if ( (nextsp = index( sp, ',' )) == NULL ) {
		if ( (nextsp = index( sp, '}' )) == NULL )
			err_msg( sp, "<- What is that?", FALSE );
	}
	
	return( nextsp );
}

/********************************************************************
 *
 *	TITLE:	   parse
 *
 *	CALLING SEQUENCE:
 *
 *		   parse( argv );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
parse( argv )
char *argv;
{
int p_num;
char *sp;
char yes;


	if ( (sp = index( argv, '=' )) != 0 )
		*sp = '\0';

	switch( lookup( params, argv ) ) { 
		case 0:	all_io = TRUE;
			return;
		case 1:	all_math = TRUE;
			return;
		case 2:	speed = D_NICKLE;
			c_update = 5;
			h_update = 300;
			return;
		case 3: speed = OH_WOW;
			c_update = 30;
			h_update = 600;
			return;
		case 4: 
			if ( sp == 0 )
				err_msg( argv, ", illegal 8087 syntax", FALSE );

			if ( (num_8087 = atoi( ++sp )) == 0)
				err_msg( argv, ", illegal number of 8087 tests", FALSE );
			return;
		case 5: if ( sp == 0 )
				err_msg( argv, ", illegal calibrate syntax", FALSE );

			if ( (cal = atol( ++sp )) == 0)
				err_msg( argv, ", illegal calibrate number", FALSE );
			c_update = 60;
			return;
		case 6:	idleps = TRUE;
			return;
		default:	;
	}


	/*
	 *  It must be a file name parameter,
	 *  so check for the left bracket.
	 */

	if ((sp = index( argv, '{' )) == 0 )
		err_msg( argv, " <- What is that?", FALSE );

	/*
	 *  Check for to many path name parameters
	 */

	if ( pi >= P_MAX )
		err_msg( "Too many file names", "", FALSE );

	/*
	 *  Save path name parameter
	 */

	*sp = '\0';
	if ( strlen( argv ) > NAMEMAX - 1 )
		err_msg( argv, ", file name too long", FALSE );
	strcpy( paths[pi].name, argv );

	/*
	 *  Find comma or end of file name parameter
	 */

	argv = sp + 1;
	if ( (sp = find( argv )) == NULL )
		err_msg( "Missing file name parameters", "", FALSE );

	*sp = '\0';

	/*
	 *  Check for optional "fill" parameter.
	 */

	if ( *argv == 'f' || *argv == 'F' ) {

		switch( lookup( fpram, argv ) ) {

					/*
					 *  Found optional "fill" parameter
					 */

			case 0:		paths[pi].number = 20;
					break;
			default:	err_msg( argv, ", illegal fill parameter", FALSE );
					break;
		}
	}
	else {
		
		/*
		 *  Parse for number of tests
		 */

		if ( (paths[pi].number = atoi( argv )) == 0 )
			err_msg( argv, ", illegal number of tests", FALSE );
	}

	/*
	 *   Initialize min and max values
	 */

	 paths[pi].min = 0xffff;
	 paths[pi].max = 0xffff;

	/*
	 *  Find comma or end of file name parameter
	 */

	argv = sp + 1;
	if ( (sp = find( argv )) != NULL ) {

		*sp = '\0';

		/*
		 *  Parse for min buffer size
		 */

		if ( (paths[pi].min = atol( argv )) == 0 )
			err_msg( argv, ", illegal min buffer size", FALSE );

		/*
		 *  Find comma or end of file name parameter
		 */

		argv = sp + 1;
		if ( (sp = find( argv )) != NULL ) {

			*sp = '\0';

			/*
			 *  Parse for max buffer size
			 */

			if ( (paths[pi].max = atol( argv )) == 0 )
				err_msg( argv, ", illegal max buffer size", FALSE );
		}
		else {
		
			/*
			 *  No max value, so use min value as buffer size
			 */
		
			paths[pi].max = paths[pi].min;
		}
	}

	/*
	 *  Check for illegal min and max values
	 */

	if ( paths[pi].min > paths[pi].max )
		err_msg( "Illegal min-max values:  min > max", "", FALSE );

	/*
	 *  Consume rest of parameter
	 */

	 if ( find( ++sp ) != NULL )
		err_msg( sp, " <- What is that?", FALSE );

	/*
	 *  Increment path index for next path name parameter
	 */

	++pi;
}

/********************************************************************
 *
 *	TITLE:	   xclock
 *
 *	CALLING SEQUENCE:
 *
 *		   xclock();
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
xclock()
{
long stime;
long ctime;
long minutes;
struct msg_dsc cmsgs;


	stime = time(0);

	for( ; ; ) {

		sleep( c_update );

		ctime = time(0) - stime;
		minutes = ctime % 3600;
		sprintf( cmsgs.msg, "%3ld:%02ld:%02ld", ctime / 3600, minutes / 60, minutes % 60 );
		xadd( &cmsgs, 22, 64 );
		xrefresh( &cmsgs );
	}
}

/********************************************************************
 *
 *	TITLE:	   idlecnt
 *
 *	CALLING SEQUENCE:
 *
 *		   idlecnt();
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
idlecnt()
{
long i;
long limit;
long b_time;
long c_time;
long e_time;
long s_time;
long o_idle;
long ot_idle;
long idle;
long tidle;
long minutes;
struct msg_dsc imsgs;


	if (cal != 0)
		limit = cal;
	else
		limit = 1898963L;

	minutes = 0;

	o_idle = -1;
	ot_idle = -1;

	s_time = time( 0 );

	for( ; ; ) {
		
		b_time = time(0);

		for ( i = 0; i <= limit; i++ )
			;

		e_time = time(0);

		if ( (c_time = e_time - b_time) > 0 ) {
			idle = 6000 / c_time;

			if ( idle != o_idle ) {
				sprintf( imsgs.msg, "%3ld", idle );
				xadd( &imsgs, 22, 50 );
				xrefresh( &imsgs );
				o_idle = idle;
			}
		}
		else {
			xaddstr( &imsgs, 22, 48, "? / 0" );
			xrefresh( &imsgs );
		}

		++minutes;
		c_time = e_time - s_time;

		if ( c_time >= 60 ) {
			tidle = (minutes * 100) / (c_time / 60);


			if ( tidle != ot_idle ) {
				sprintf( imsgs.msg, "%3ld", tidle );
				xadd( &imsgs, 22, 57 );
				xrefresh( &imsgs );
				ot_idle = tidle;
			}
		}
		else {
			xaddstr( imsgs.msg, 22, 55, "? / 0" );
			xrefresh( &imsgs );
		}
	}
}

/********************************************************************
 *
 *	TITLE:	   screen
 *
 *	CALLING SEQUENCE:
 *
 *		   screen();
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
screen()
{
struct msg_dsc msgs;

	nice(1);

	for( ; ; ) {
	
		xread( &msgs );
		
		switch( msgs.code ) {
			case REFCODE:	refresh();
					break;
			case STRCODE:	mvaddstr( msgs.y, msgs.x, msgs.msg );
					break;
			case FCODE:	standout();
					mvaddstr( msgs.y, msgs.x, msgs.msg );
					standend();
					break;
			case XREFCUR:	werase( curscr );
					touchwin( stdscr );
					refresh();
					break;
		}

	}
}

/********************************************************************
 *
 *	TITLE:	   main
 *
 *	CALLING SEQUENCE:
 *
 *		   main( argc, argv );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
main( argc, argv )
int argc;
char *argv[];
{
int	i;
int time;
int status;	 


	/*
	 *  Set up abort procedure
	 */

	signal( SIGINT, quit );

	/*
	 *  Initialization
	 */

	for (i = 0; i < P_MAX; ++i )
		strcpy( paths[i].name, "" );

	/*
	 *  Parse command line
	 */

	for (i = 1; i < argc; i++ )
		parse( argv[i] );

	/*
	 *  Check if user has mixed up math and io parameters
	 */

	if ( (all_io && all_math) || (all_math && pi != 0) || (all_io && num_8087 != 0) )
		err_msg( "Illegal mix of IO and MATH parameters", "", FALSE );

	/*
	 *  Initialize semaphore pipe
	 */

	sema();

	/*
	 *  Initialize screen pipe
	 */

	xpipe();

	/*
	 *  Create pipes for process synchronization
	 */

	status = pipe( synch );
	if ( status == E_IO )
		err_msg( "Bad parent create of synch pipe", "", TRUE );

	status = pipe( go );
	if ( status == E_IO )
		err_msg( "Bad parent create of go pipe", "", TRUE );

	/*
	 *  Initialize random number generator
	 */

	dice();

	/*
	 *  Initialize Curses
	 */

	initcio();
	cur_on = TRUE;

	if (cal == 0) {

		/*
		 *  Generate tests
		 */

		 gen16();

		/*
		 *  Generate headers
		 */

		genhead();

		/*
		 *  Create repaint process
		 */

		forkret = fork();
		if ( forkret == -1 )
			err_msg( "Bad fork of repaint process", "", TRUE );
		else {
			if ( forkret == 0 ) {
				signal( SIGINT, SIG_DFL );
				repaint();
			}
		}
 	
		/*
		 *  Wait a couple of seconds for headers
		 *  then signal tests to start
		 */

		sleep( 2 );
		status = write( go[1], go_msg, strlen( go_msg ) );
		if (status == E_IO)
			err_msg( "Bad parent write of go pipe", "", TRUE );
 	}

	/*
	 *  Create Clock
	 */

	forkret = fork();
	if ( forkret == -1 )
		err_msg( "Bad fork of clock process", "", TRUE );
	else {
		if ( forkret == 0 ) {
			signal( SIGINT, SIG_DFL ); 
			xclock();
		}
	}


	/*
	 *  Create idle process
	 */
	if ( idleps ) {

		forkret = fork();
		if ( forkret == -1 )
			err_msg( "Bad fork of idle process", "", TRUE );
		else {
			if ( forkret == 0 ) {
				signal( SIGINT, SIG_DFL );

				/*
				 *  Drop priority by 20
				 */

				nice(20);
				idlecnt();
			}
		}
	}
 	
	/*
	 *  Go handle the screen for everbody
	 */
	 
	 screen();
 }

